from decimal import Decimal
from flask import Flask, request, jsonify
from qval import validate, qval_curry, Validator
from qval.framework_integration import setup_flask_error_handlers

app = Flask(__name__)

# Flask uses one global request object.
# You can automatically provide it to the `@qval()` decorator on the each call
# by executing `qval_curry(request)`.
qval = qval_curry(request)

# Setup exception handlers
setup_flask_error_handlers(app)


@app.route("/api/divide")
def division_view():
    """
    GET /api/divide?
    param a : int
    param b : int, nonzero

    Example: GET /api/divide?a=10&b=2&token=abcdefghijkl -> 200, {"answer": 5}
    """
    # Parameter validation occurs in the context manager.
    # If validation fails or user code throws an error, context manager
    # will raise InvalidQueryParamException or APIException respectively.
    # In Django, these exception will be processed and result
    # in error codes (400 and 500) on the client side.
    params = (
        validate(request, a=int, b=int)
        # `b` must be anything but zero
        .nonzero("b")
    )
    with params as p:
        return jsonify({"answer": p.a // p.b})


@app.route("/api/pow")
@qval({"a": float, "b": float})
def pow_view(request, params):
    """
    Note that request parameter is now being provided to the view.

    GET /api/pow?
    param a : float
    param b : float

    Example: GET /api/pow?a=2&b=3 -> 200, {"answer": 8.0}
    Example: GET /api/pow?a=2&b=3000000000000 -> 500,
    {
        "error": "An error occurred while processing you request. Please contact the website administrator."
    }
   """
    # Here we don't catch the OverflowError if `b` is too big.
    # This will result in 500 error on the client side.
    return jsonify({"answer": params.a ** params.b})


purchase_factories = {"price": Decimal, "item_id": int, "token": None}
purchase_validators = {
    "price": Validator(lambda x: x > 0),
    "token": Validator(lambda x: len(x) == 12),
    "item_id": Validator(lambda x: x >= 0),
}


@app.route("/api/purchase")
@qval(purchase_factories, purchase_validators)
def purchase_view(request, params):
    """
    GET /api/purchase?
    param item_id : int, positive
    param price   : float, greater than zero
    param token   : string, length == 12

    Example: GET /api/purchase?item_id=1&price=5.8&token=abcdefghijkl
             -> {"success": "Item '1' has been purchased. Check: 5.92$."
    """
    tax = 0.02
    cost = params.price * Decimal(1 + tax)
    return jsonify({
        "success": f"Item '{params.item_id}' has been purchased. Check: {round(cost, 2)}$."
    })



"""
This module provides convenient API for verifying query parameters.

The core class is called `QueryParamValidator`. It accepts 4 arguments:
- request: Request instance (Any object that has following attributes: GET, query_params and body)

- factories: Dictionary of factories {param -> factory}. The value of the parameter will be provided
  to factory before validation. Any callable that accepts string and returns anything is a valid factory.

- validators: Dictionary if validators {param -> Validator}. Validator is basically a list of predicates
  with the __call__() operator. See `Validator` class for more info.

- box_all: If true, adds all request parameter to the final result.
  Otherwise, only specified in `factories` parameters will be added.

If any parameter fails validation, `InvalidQueryParamException` (HTTP code = 400) will be raised.
Also, only `TypeError`, `ValueError` and `KeyError` occurred when param is provided to factory
result in the same exception.
Any error thrown inside or outside of the context will raise an APIError (HTTP code = 500).

Example:
    >>> from qval.framework_integration import DummyRequest
    >>> r = DummyRequest({"num": "42"})
    >>> with QueryParamValidator(r, {"num": int}) as p:
    ...     print(p.num)
    42

The code above is to verbose. That's why you should use `validator()` -
this function does the all boilerplate work for you:
    - `validator()` automatically converts dictionary to Request-like objects
    - Key-value arguments are used to provide factories
    - It's easier to type

    Simple example:
    >>> r = {"num": "42", "string": "s", "price": "3.14"}  # you can use dictionary instead of Request instance
    >>> with validate(r, num=int, price=float) as p:
    ...     print(p.num, p.price, p.string, sep=', ')
    42, 3.14, s

    A little bit more complex, with custom factory:
    >>> r = {"price": "2.79$", "tax": "0.5$"}
    >>> currency2f = lambda x: float(x[:-1])  # factory that converts {num}$ to float
    >>> with validate(r, price=currency2f, tax=currency2f) as p:
    ...     print(p.price, p.tax, sep=', ')
    2.79, 0.5


You can also use `qval()` decorator:
    >>> factories = {"num": int, "special": float}
    >>> validators = {"special": Validator(lambda x: x > 0)}
    >>> @qval(factories, validators)
    ... def view(request, params):  # class-based views are also supported
    ...     print(params.num, params.special, sep=", ")
    >>> view({"num": "10", "special": "0.7"})
    10, 0.7

    If something fails during validation or inside of the function, an error will be thrown.
    Consider the following examples:
    >>> factories = {"num": int, "special": int}  # now special is an integer
    >>> @qval(factories, validators=None)  # no validators for simplicity
    ... def view(request, params):
    ...     pass
    >>> view({"num": "10", "special": "0.7"})  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
        ...
    qval.exceptions.InvalidQueryParamException:
        {'error': 'Invalid type of the `special` parameter: expected int.'}.

    The HTTP code of the exception above is 400 (Bad Request).

    Now the error is raised inside of the context block:
    >>> factories = {"num": int, "special": float}
    >>> @qval(factories, validators=None)  # no validators for simplicity
    ... def view(request, params):
    ...     raise IOError  # some random exception
    >>> view({"num": "10", "special": "0.7"})  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
        ...
    qval.drf_integration.APIException: An error occurred while processing you request.
                                       Please contact website administrator.

    The HTTP code of the exception above is 500 (Internal Server Error).
    The error is logged to stdout by default. See the Note section for more info.

Notes:
    TODO: add notes
"""
from .utils import log
from .qval import QueryParamValidator, validate, qval, qval_curry
from .exceptions import InvalidQueryParamException, APIException
from .validator import Validator


__version__ = "0.1.5"



import pytest
from qval import validate, APIException, InvalidQueryParamException
from qval.framework_integration import (
    HTTP_500_INTERNAL_SERVER_ERROR,
    HTTP_400_BAD_REQUEST,
)


def test_params_processed():
    dct = {"num": "42", "double": "2.79", "string": "some string"}
    with validate(dct, num=int, double=float) as p:
        assert p.num == 42
        assert p.double == 2.79
        assert p.string == "some string"


def test_params_omitted():
    dct = {"num": "42", "string": "some string"}
    # Disable auto-detection of parameters (box_all)
    params = validate(dct, num=int, box_all=False)
    with pytest.raises(APIException) as e, params as p:
        assert p.num == 42
        assert p.string == "some string"
    assert e.value.status_code == HTTP_500_INTERNAL_SERVER_ERROR


def test_missing_param_throws_error():
    dct = {"param1": "whatever", "param2": "6.66"}
    params = validate(dct, param1=None, param2=float, param3=int)
    with pytest.raises(InvalidQueryParamException) as e, params:
        pass
    assert e.value.status_code == HTTP_400_BAD_REQUEST


def test_validator_factory():
    currency2f = lambda x: float(x[:-1])
    r = {
        "price": "43.5$",
        "n_items": "1",
        "meta": "info",
        "num": 10,
        "num2": 5,
        "token": "0123456789",
    }
    params = (
        validate(
            r, price=currency2f, n_items=int, num=int, num2=int
        ).positive(  # `n_items` must be greater than zero
            "n_items"
        )
        # `num` must be equal to 10
        .eq("num", 10)
        # `num2` must be less than 10
        .lt("num2", 10)
        # Len of `token` must be equal to 10
        .check("token", lambda x: len(x) == 10)
        # The same check as above, but using `transform`
        .eq("token", 10, transform=len)
    )
    with params as p:
        assert {43.5, 1, "info", 10, 5, "0123456789"} == set(p.__dct__.values())


def test_validation_fails():
    currency2f = lambda x: float(x[:-1])
    r = {
        "price": "43.5$",
        "n_items": "0",
        "meta": "info",
        "num": -10,
        "num2": 20,
        "token": "012345678",
    }
    params = (
        validate(
            r, price=currency2f, n_items=int, num=int, num2=int
        ).positive(  # `n_items` must be greater than zero
            "n_items"
        )
        # `num` must be equal to 10
        .eq("num", 10)
        # `num2` must be less than 10
        .lt("num2", 10)
        # Len of `token` must be equal to 10
        .check("token", lambda x: len(x) == 10)
        # The same check as above, but using `transform`
        .eq("token", 10, transform=len)
    )
    with pytest.raises(InvalidQueryParamException) as e, params:
        pass
    assert e.value.status_code == HTTP_400_BAD_REQUEST


def test_exception_handled_in_outside_context():
    """
    See `QueryParamValidator._validate()` and `test_supported_errors_handled()` for more info.
    """
    # Random exception.
    def f(_):
        raise IOError

    r = {"param": "value"}
    with pytest.raises(APIException) as e, validate(r, param=f):
        pass
    assert e.value.status_code == HTTP_500_INTERNAL_SERVER_ERROR


def test_supported_errors_handled():
    """
    Only TypeError, ValueError and KeyError that occurred during the validation
    are handled as expected. Any error thrown inside of the context will raise APIError.
    See `test_unsupported_errors_handled()`.
    """

    def exc_factory(exc):
        def f(_):
            raise exc

        return f

    r = {"param": "value"}
    for exc in (TypeError, ValueError, KeyError):
        with pytest.raises(InvalidQueryParamException) as e, validate(
            r, param=exc_factory(exc)
        ):
            pass
        assert e.value.status_code == HTTP_400_BAD_REQUEST


def test_unsupported_errors_handled():
    supported_exceptions = (TypeError, ValueError, KeyError)
    random_exceptions = (IOError, BrokenPipeError, ConnectionError, BufferError)
    r = {"param": "value"}
    for exc in supported_exceptions + random_exceptions:
        with pytest.raises(APIException) as e, validate(r):
            raise exc
        assert e.value.status_code == HTTP_500_INTERNAL_SERVER_ERROR



from decimal import Decimal
from qval import Validator

purchase_factories = {"price": Decimal, "item_id": int, "token": None}

purchase_validators = {
    "price": Validator(lambda x: x > 0),
    "token": Validator(lambda x: len(x) == 12),
    "item_id": Validator(lambda x: x >= 0),
}



import pytest
from decimal import Decimal

from qval import InvalidQueryParamException, qval, qval_curry
from qval.framework_integration import HTTP_400_BAD_REQUEST, Request
from qval.utils import FrozenBox
from qval.validator import Validator


stats = []
factories = {
    "double": float,
    "num": int,
    "price": Decimal,
    "hashme": hash,
    "observable": stats.append,
}
validators = {"price": Validator(lambda x: x > 0)}


@qval(factories, validators)
def simple_view(request: Request, params: FrozenBox):
    return request, params


class ViewClass(object):
    @qval(factories, validators)
    def complex_view(
        self, request: Request, param1: int, param2: int, params: FrozenBox
    ):
        return request, param1, param2, params


def test_params_provided():
    request = {
        "double": "3.14",
        "num": "10",
        "price": "2.79",
        "hashme": "s$cret_t0k3n",
        "observable": "important metric",
    }
    # Test simple view
    r, params = simple_view(request)
    assert stats[-1] == request["observable"]
    assert set(r.query_params.keys()) == set(params.__dct__.keys())

    # Test complex view
    r, p1, p2, params = ViewClass().complex_view(request, 1, 2)
    assert (p1, p2) == (1, 2)
    assert stats[-1] == request["observable"]
    assert set(r.query_params.keys()) == set(params.__dct__.keys())


def test_params_validated():
    request = {
        "double": "3.14",
        "num": "10",
        "price": "0",
        "hashme": "s$cret_t0k3n",
        "observable": "important metric",
    }
    with pytest.raises(InvalidQueryParamException) as e:
        simple_view(request)
    assert stats[-1] == request["observable"]
    assert e.value.status_code == HTTP_400_BAD_REQUEST

    with pytest.raises(InvalidQueryParamException) as e:
        ViewClass().complex_view(request, 1, 2)
    assert stats[-1] == request["observable"]
    assert e.value.status_code == HTTP_400_BAD_REQUEST


# Curries qval with a static request (in this case)
# Useful in frameworks with global request classes (e.g. Flask)
def get_curried_qval():
    request = {
        "double": "3.14",
        "num": "10",
        "price": "2.79",
        "hashme": "s$cret_t0k3n",
        "observable": "important metric",
    }
    return qval_curry(request)

curried_qval = get_curried_qval()


@curried_qval(factories, validators)
def view(request, some_param, params):
    return request, some_param, params


def test_curried_qval():
    # Test simple view
    r, test, params = view("test")
    assert test == "test"
    assert stats[-1] == r.query_params["observable"]
    assert set(r.query_params.keys()) == set(params.__dct__.keys())



import logging
from pprint import pformat
from typing import Dict, Any, List, Callable, Union

from . import framework_integration as fwk


@fwk._make_request
def make_request(request: Union[Dict[str, str], fwk.Request]) -> fwk.Request:
    """
    Creates DummyRequest if `request` is dictionary, and returns the `request` itself otherwise.
    :param request: dict or request instance
    :return: request
    """
    if isinstance(request, dict):
        return fwk.DummyRequest(request)
    return request


class FrozenBox(object):
    """
    Frozen dictionary that allows access to elements by `.`.
    """

    def __init__(self, dct: Dict[Any, Any]):
        """
        :param dct: dict to store
        """
        self.__dct__ = dct

    def __getitem__(self, item: str) -> Any:
        """
        [] operator.

        :param item: item key
        :return: value for key `item`
        """
        return self.__dct__[item]

    def __getattr__(self, item: str) -> Any:
        """
        Returns value of the stored `item` or attribute of the object.

        :param item: item key
        :return: value
        """
        if item == "__dct__":
            return getattr(super(), item)
        return self[item]

    def __contains__(self, item: str) -> bool:
        """
        Determines if item is inside of the dictionary.
        :param item: item to check
        """
        return item in self.__dct__

    def __iter__(self):
        """
        Returns iterator over self.__dct__.values()
        :return:
        """
        return iter(self.__dct__)

    def __repr__(self) -> str:
        """
        Returns evaluable representation of the FrozenBox object.
        """
        return f"FrozenBox({self.__dct__})"

    def __str__(self) -> str:
        """
        Returns string representation of the FrozenBox object.
        :return: str(box)
        """
        return f"FrozenBox<{self.__dct__}>"


class ExcLogger(object):
    """
    A class used in query parameters validation to report critical errors.
    """

    def __init__(self, logger_factories: List[Callable[[str], Any]]):
        """
        Instantiates the ExcLogger.
        :param logger_factories: list of logger factories
        """
        self.factories = logger_factories
        self._enabled = True

    @property
    def is_enabled(self) -> bool:
        """
        Returns True if logging is enabled.
        """
        return self._enabled

    def disable(self):
        """
        Disables logging.
        :return: None
        """
        self._enabled = False

    def enable(self):
        """
        Enables logging.
        :return: None
        """
        self._enabled = True

    def add_logger(self, log_factory: Callable[[str], Any]):
        """
        Adds new logger factory to list.

        :param log_factory: logger
        :return: None
        """
        self.factories.append(log_factory)

    def clear(self):
        """
        Removes all saved factories.
        :return: None
        """
        self.factories.clear()

    def error(self, name: str, *args, **kwargs):
        """
        Creates new logger using provided factories and dumps error message.

        :param name: logger name
        :param args: logger args
        :param kwargs: logger kwargs
        :return: None
        """
        if not self._enabled:
            return

        for log in self.factories:
            try:
                log(name).error(*args, **kwargs)
            except:
                pass

    @classmethod
    def detect_loggers(cls) -> "ExcLogger":
        """
        Looks for configuration and instantiates ExcLogger with detected loggers
        or default logging.getLogger
        :return: ExcLogger object
        """
        module = fwk.get_module()
        if hasattr(module, "QVAL_LOGGERS"):
            loggers = module.QVAL_LOGGERS
        else:
            loggers = [logging.getLogger]

        return cls(loggers)

    def __repr__(self) -> str:
        return f"ExcLogger([{', '.join(map(lambda x: x.__name__, self.factories))}])"

    def __str__(self) -> str:
        factories = pformat(str(self.factories)).strip("'")
        return f"ExcLogger({factories}, enabled = {self.is_enabled})"


# Detect loggers
log = ExcLogger.detect_loggers()

# Remove ExcLogger, this will make `log` acting as SingleTon
del ExcLogger



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['qval']

package_data = \
{'': ['*']}

setup(name='qval',
      version='0.1.5',
      description='This module provides convenient API for verifying query parameters.',
      author='George Usan-Podgornov',
      author_email='',
      url='https://github.com/OptimalStrategy/Qval',
      packages=packages,
      package_data=package_data,
     )



import logging
import os
from typing import Union, Dict
from importlib import import_module


class _EnvironSettings(object):
    """
    Lookups attribute calls in os.environ.
    """

    def __getattr__(self, item):
        item = os.environ.get(item)
        # Support `hasattr()`
        if item is None:
            raise AttributeError
        return item


class DummyRequest(object):
    """
    DummyRequest. Used for compatibility with frameworks.
    """

    def __init__(self, params: Dict[str, str]):
        self.GET = params
        self.body = "<DummyRequest: no body>"

    @property
    def query_params(self) -> Dict[str, str]:
        """
        More semantically correct name for request.GET.
        """
        return self.GET


Request = DummyRequest


def get_module() -> Union[_EnvironSettings, "Module"]:
    """
    Attempts to load settings module.
    If none of the supported env variables are defined, returns `_EnvironSettings()` object.
    """
    module = None
    modules = ["DJANGO_SETTINGS_MODULE", "SETTINGS_MODULE"]
    for module in map(os.environ.get, modules):
        if module is not None:
            module = module.replace(".py", "").replace("/", ".")
            break
    return _EnvironSettings() if module is None else import_module(module)


module = get_module()


def load_symbol(path: Union[object, str]):
    """
    Imports object using the given path.

    :param path: path to an object, e.g. my.module.func_1
    :return: loaded symbol
    """
    # Path is already a symbol
    if not isinstance(path, str):
        return path
    _mod, _symbol = path.rsplit(".", maxsplit=1)
    return getattr(import_module(_mod), _symbol)


# Check if DRF is installed
try:
    from rest_framework.request import Request as _Request
    from rest_framework.exceptions import APIException
    from rest_framework.status import (
        HTTP_400_BAD_REQUEST,
        HTTP_500_INTERNAL_SERVER_ERROR,
    )

    Request = _Request
    REST_FRAMEWORK = True
except ImportError:
    REST_FRAMEWORK = False

    # Define missing symbols
    class APIException(Exception):
        def __init__(self, detail: Union[dict, str]):
            self.detail = detail
            self.status_code = HTTP_500_INTERNAL_SERVER_ERROR
            super().__init__(detail)

    HTTP_400_BAD_REQUEST = 400
    HTTP_500_INTERNAL_SERVER_ERROR = 500

    if hasattr(module, "QVAL_REQUEST_CLASS"):
        Request = load_symbol(module.QVAL_REQUEST_CLASS)


# Check if Django is installed
try:
    from django.http import HttpRequest, JsonResponse

    # Exit if DRF is installed
    if REST_FRAMEWORK:
        raise ImportError

    Request = HttpRequest

    class HandleAPIExceptionDjango(object):
        def __init__(self, get_response):
            self.get_response = get_response

        def __call__(self, request):
            return self.get_response(request)

        def process_exception(self, _: Request, exception: Exception):
            if isinstance(exception, APIException):
                detail = exception.detail
                if isinstance(detail, str):
                    detail = {"error": detail}
                return JsonResponse(detail, status=exception.status_code)

    if hasattr(module, "MIDDLEWARE"):
        module.MIDDLEWARE.append("qval.framework_integration.HandleAPIExceptionDjango")
    else:
        logging.warning(
            "Unable to add APIException middleware to the MIDDLEWARE list. "
            "Django does not support APIException handling without DRF integration. "
            "Define DJANGO_SETTINGS_MODULE or add 'qval.framework_integration.HandleAPIExceptionDjango' "
            "to the MIDDLEWARE list."
        )
except ImportError:
    pass


# Check if custom wrapper is provided
if hasattr(module, "QVAL_MAKE_REQUEST_WRAPPER"):
    _make_request = load_symbol(module.QVAL_MAKE_REQUEST_WRAPPER)
else:

    def _make_request(f):
        """
        Wraps default `utils.make_request()` function. Does nothing.
        """
        return f


def setup_flask_error_handlers(app: "flask.Flask"):
    """
    Setups error handler for APIException.

    :param app: flask app
    :return: None
    """
    from flask import jsonify

    @app.errorhandler(APIException)
    def handle_api_exception(error: APIException):
        response = error.detail
        if isinstance(response, str):
            response = {"error": response}
        response = jsonify(response)
        response.status_code = error.status_code
        return response
